import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.PrintWriter;
import java.io.IOException;
import java.net.Socket;
import java.util.concurrent.Semaphore;
import javax.swing.JTextField;
import javax.swing.text.StyledDocument;

/**
 * Ein Sendeprozess fr den Spielclient.
 * 
 * @author Q11
 * @version 2015-06-16
 */
public class ClientSendeProzess implements Runnable, ActionListener, KeyListener, MouseListener
{
    private Socket meinSocket;
    private PrintWriter gibAnServer;
    private Spiel meinSpiel;
    private JTextField meinEingabeFeld;
    private StyledDocument meinAusgabeDokument;
    private Semaphore aufhoeren;

    /**
     * Konstruktor.
     * Verbinde Dich an den angegebenen Socket.
     */
    public ClientSendeProzess (Socket s)
    { 
        meinSocket          = s;
        meinEingabeFeld     = Chatfenster.gibChatfenster ().gibEingabe ();
        meinAusgabeDokument = Chatfenster.gibChatfenster ().gibAusgabe ().getStyledDocument ();

        aufhoeren = new Semaphore (1);
    }

    /**
     * Ordne mir ein neues Spiel zu.
     */
    public void setzeSpiel (Spiel neuesSpiel)
    {
        if (meinSpiel != null)
        {
            Leinwand.gibLeinwand ().gibFenster ().removeKeyListener (this);
            Leinwand.gibLeinwand ().gibZeichenflaeche ().removeMouseListener (this);
            Chatfenster.gibChatfenster ().gibEingabe ().removeActionListener (this);
        }

        meinSpiel = neuesSpiel;

        Leinwand.gibLeinwand ().gibFenster ().addKeyListener (this);
        Leinwand.gibLeinwand ().gibZeichenflaeche ().addMouseListener (this);
        Chatfenster.gibChatfenster ().gibEingabe ().addActionListener (this);
    }

    /**
     * Ablaufplan.
     * Sende die Tastatureingabe an den Server, bis sie "quit" lautet.
     */
    public void run ()
    { 
        try 
        {
            /*
             * Datenstrome an den Server.
             */
            gibAnServer     = new PrintWriter (meinSocket.getOutputStream (), true);

            /*
             * Endesignal.
             */
            aufhoeren.drainPermits ();

            /*
             * Chatte, solange die Benutzerin nicht das Endesignal eingibt.
             */
            try
            {
                aufhoeren.acquire ();
            }
            catch (Exception fehler)
            {
            }

            /*
             * Schliee den Datenstrom.
             */
            gibAnServer.close ();

            /*
             * Beende das Auslesen des Eingabefelds.
             */
            meinEingabeFeld.removeActionListener (this);
        }
        catch (IOException fehler) 
        {
            System.err.println ("Keine Verbindung beim Senden.");
        }
    }

    /**
     * Sende eine Nachricht an den Server.
     */
    private void sende (String nachricht)
    {
        if (0 == aufhoeren.availablePermits ())
        {
            gibAnServer.println (nachricht);

            try
            {
                meinAusgabeDokument.insertString (meinAusgabeDokument.getLength (), nachricht + "\n", Chatfenster.gibChatfenster ().gibEingabestil ());
            }
            catch (Exception fehler)
            {
            }

            if (nachricht.equals ("quit"))
            {
                aufhoeren.release ();
            }
        }
    }

    /**
     * Sende eine Nachricht an den Server aber nicht ans Chatfenster.
     */
    private void sendeStill (String nachricht)
    {
        if (0 == aufhoeren.availablePermits ())
        {
            gibAnServer.println (nachricht);

            if (nachricht.equals ("quit"))
            {
                aufhoeren.release ();
            }
        }
    }

    /**
     * Reagiere auf Aktionen in den beobachteten Komponenten.
     */
    public void actionPerformed (ActionEvent ereignis)
    {
        if (ereignis.getSource () == meinEingabeFeld)
        {
            sende (meinEingabeFeld.getText ());

            meinEingabeFeld.setText ("");
        }
    }

    /**
     * Reagiere nicht auf das Drcken einer Taste.
     */
    public void keyPressed (KeyEvent taste)
    {   
    }

    /**
     * Reagiere auf das Tippen einer darstellbaren Taste.
     */
    public void keyTyped (KeyEvent taste)
    {
        char zeichen = taste.getKeyChar ();

        try
        {
            switch (zeichen)
            {
                case 'r':
                sende ("resign");
                break;

                case 'd':
                sende ("done");
                break;

                default:
                sendeStill ("move " + zeichen);
                break;
            }
        }
        catch (Exception fehler) 
        {
            System.err.println ("Keine Verbindung beim Senden.");
        }
    }

    /**
     * Reagiere nicht auf das Loslassen einer Taste.
     */
    public void keyReleased (KeyEvent taste)
    {
    }

    /**
     * Reagiere nicht auf das Druecken der Maustaste.
     */
    public void mousePressed (MouseEvent ereignis)
    {
    }

    /**
     * Reagiere auf das Loslassen des Mausknopfes.
     */
    public void mouseReleased (MouseEvent ereignis)
    {
        int qx = meinSpiel.gibQuerlinieX (ereignis.getPoint ());
        int qy = meinSpiel.gibQuerlinieY (ereignis.getPoint ());

        if (qx >= 0 && qy >= 0)
        {
            sendeStill ("move quer " + qx + " " + qy);
        }

        int hx = meinSpiel.gibHochlinieX (ereignis.getPoint ());
        int hy = meinSpiel.gibHochlinieY (ereignis.getPoint ());

        if (hx >= 0 && hy >= 0 )
        {
            sendeStill ("move hoch " + hx + " " + hy);
        }
    }

    /**
     * Reagiere nicht auf das Klicken mit der Maus.
     */
    public void mouseClicked (MouseEvent ereignis)
    {
    }

    /**
     * Reagiere nicht auf das Verlassen der Leinwand mit der Maus.
     */
    public void mouseExited (MouseEvent ereignis)
    {
    }

    /**
     * Reagiere nicht auf das Betreten der Leinwand mit der Maus.
     */
    public void mouseEntered (MouseEvent ereignis)
    {
    }
}